2 ICON "src/icons/horizon-eda.ico"

1 VERSIONINFO
FILEVERSION     0,0,0,0
PRODUCTVERSION  0,0,0,0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
      VALUE "FileDescription", "Horizon project manager"
      VALUE "InternalName", "horizon-prj-mgr"
      VALUE "OriginalFilename", "horizon-prj-mgr.exe"
      VALUE "ProductName", "horizon EDA"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x809, 1252
  END
END